
// (c) COPYRIGHT URI/MIT 1996,1997,1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

// Implementation of the Gui class. A GUI is a program runing in a subprocess
// which reads commands from its stdin and writes responses to stdout. The
// GUI presents inforamtion to the user using the usual combination of
// graphical candy.
//
// jhrg 6/21/96

// $Log: Gui.cc,v $
// Revision 1.14  1999/04/29 02:29:30  jimg
// Merge of no-gnu branch
//
// Revision 1.13.14.2  1999/02/05 09:32:34  jimg
// Fixed __unused__ so that it not longer clashes with Red Hat 5.2 inlined
// math code.
//
// Revision 1.13.14.1  1999/02/02 21:56:59  jimg
// String to string version
//
// Revision 1.13  1997/06/09 20:58:47  jimg
// Added `etc/' to default GUI initialization command location
// (See _start_gui()).
//
// Revision 1.12  1997/04/18 17:26:22  jimg
// Changed include of String.h so that it uses `<>' quotes instead of `""'
// quotes since the latter messed up the dependencies in the Makefile.
//
// Revision 1.11  1997/03/05 08:26:51  jimg
// Fixed bug with _show_gui -vs- _user_wants_gui fields.
//
// Revision 1.10  1997/02/19 04:51:58  jimg
// Fixed bugs with _show_gui and _user_wants_gui.
//
// Revision 1.9  1997/02/10 02:34:40  jimg
// Changed _start_gui() call to dods_root to match new return type.
//
// Revision 1.8  1996/11/25 03:41:28  jimg
// Fixed dods_root so that it now uses the function (rather that a static
// global which won't be initialized when this library is linked using a
// program other than g++).
//
// Revision 1.7  1996/11/22 00:06:23  jimg
// Increased Expect timeout from 2 to 10 seconds.
//
// Revision 1.6  1996/11/18 21:59:32  jimg
// Error messages from command are more verbose - to simplify debugging.
//
// Revision 1.5  1996/11/13 18:58:40  jimg
// Added missing break statements to switch()s.
//
// Revision 1.4  1996/08/13 18:27:46  jimg
// Added not_used in definition of char rcsid[].
//
// Revision 1.3  1996/07/08 23:19:16  jimg
// Added support for the HAVE_EXPECT compile-time switch. When defined as 1
// support for expect is built, when 0 it is not. Thus the dap++ library may be
// built so that expect is not required to link programs. However, the GUI will
// work only when expect/tcl are present and when client are linked with those
// libraries in addition to libdap++.a
// Fixed comments.
// Fixed some warnings (but left some...).
//
// Revision 1.2  1996/06/22 00:04:23  jimg
// Removed forced DEBUG define.
//
// Revision 1.1  1996/06/21 23:12:29  jimg
// Created based on code that was in Connect.cc.
//

#ifdef __GNUG__
#pragma "implemenation"
#endif

#include "config_dap.h"

static char rcsid[] not_used = {"$Id: Gui.cc,v 1.14 1999/04/29 02:29:30 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>

#include <sys/wait.h>

#if HAVE_EXPECT
#include <expect.h>
#endif

#include <string>

#include "util.h"
#include "debug.h"
#include "Gui.h"

// Constants used to talk to the GUI.

static const int OK = 1;
static const int ERROR = 2;
static const int CANCEL = 3;
#if HAVE_EXPECT
static const char *GUI_PROMPT = "% ";
static const char *GUI_EXIT_COMMAND = "exit\r";
static const char *dods_gui = "dods_gui.tcl\r";
#endif

bool
Gui::_start_gui()
{    
    if (!show_gui())		// Don't start it if it won't be used.
	return false;

#if HAVE_EXPECT

#if DODS_DEBUG
    exp_loguser = 1;		// Defined in expect_comm.h (see expect.h).
#else
    exp_loguser = 0;
#endif

    // As per the rules, first try dods_root/etc/GUI then look on the user's
    // PATH. If both fail, return false. However, if _CMD is an absolute
    // path, don't bother prepending DODS_ROOT.

    string gui_cmd;
    if (_cmd[0] == '/')
	gui_cmd = _cmd;
    else
	gui_cmd = (string)dods_root() + "/etc/" + _cmd;

    _gui = exp_spawnl((char *)gui_cmd.c_str(), 
		      (char *)_cmd.c_str(), NULL);
    if (_gui == -1) {
	if ((_gui = exp_spawnl((char *)_cmd.c_str(), 
			       (char *)_cmd.c_str(), NULL)) == -1) {
	    _show_gui = false;
	    return false;
	}
    }
    
    _gui_pid = exp_pid;		// Save the PID so we can clean up later. 

    int status = exp_expectl(_gui, 
			     exp_exact, GUI_PROMPT, OK, 
			     exp_end);
    switch (status) {
      case OK:
	DBG(cerr << "Started the GUI" << endl);
	break;
      case ERROR:
      case EXP_TIMEOUT:
      case EXP_EOF:
      case EXP_FULLBUFFER:
      default:
	cerr << "Could not start the GUI." << endl;
	if (kill(_gui_pid, SIGKILL) == 0)
	    waitpid(_gui_pid, NULL, 0);
	_show_gui = false; // No GUI if we cannot get it working!
	_gui = -1;
	return false;
    }

    // GUI init command either comes from an environment variable or uses the
    // default (my tcl/tk program).
    const char* g = getenv("DODS_GUI_INIT");
    string gui_init = (g ? g : "");
    if (gui_init == "none")
	return true;
    else if (gui_init == "")
	gui_init = (string)"source " + dods_root() + "/etc/" + dods_gui;

    if (!command(gui_init)) {
	if (kill(_gui_pid, SIGKILL) == 0)
	    waitpid(_gui_pid, NULL, 0);
	_show_gui = false;
	_gui = -1;
	return false;
    }

    return true;

#else // !HAVE_EXPECT

    return false;

#endif
}

// Don't start the GUI, just fill in the state variables. Later if the GUI
// is pressed into service, start it then.

Gui::Gui(const string cmd) 
    : _gui(-1), _gui_pid(-1), _show_gui(true), _progress_visible(false),
      _cmd(cmd)
{
    const char* g = getenv("DODS_USE_GUI");
    string dods_gui = (g ? g : "");
    downcase(dods_gui);
    if (dods_gui == "no")
	_user_wants_gui = false; // Users explicitly say "no"...
    else
	_user_wants_gui = true; // Otherwise assume "yes".
}

Gui::Gui() 
    : _gui(-1), _gui_pid(-1), _show_gui(true), _progress_visible(false)
{
    const char* g = getenv("DODS_GUI");
    _cmd = (g ? g : "");
    if (_cmd == "")
	_cmd = "wish";

    g = getenv("DODS_USE_GUI");
    string dods_gui = (g ? g : "");
    downcase(dods_gui);
    if (dods_gui == "no")
	_user_wants_gui = false; // Users explicitly say "no"...
    else
	_user_wants_gui = true; // Otherwise assume "yes".
}

Gui::~Gui()
{
    if (!show_gui() || _gui < 0) // Don't try to stop what is not running.
	return;
    
#if HAVE_EXPECT

    string command = GUI_EXIT_COMMAND;
    unsigned status = write(_gui, command.c_str(), command.length());
    if (status != command.length()) {
	cerr << "Wrong number of chars transferred to GUI!" << endl;
	return;
    }
    
    status = exp_expectl(_gui, exp_exact, "", OK, exp_end);

    switch (status) {
      case OK:
	return;
      case EXP_TIMEOUT:
      case EXP_EOF:
      case EXP_FULLBUFFER:
      default:
	DBG(cerr << "Communication breakdown!" << endl);
	if (kill(_gui_pid, SIGKILL) == 0)
	    waitpid(_gui_pid, NULL, 0);
	_user_wants_gui = false;
	_gui = -1;
	return;
    }

    waitpid(_gui_pid, NULL, 0);

    _gui = -1;

#else // !HAVE_EXPECT

    _gui = -1;

#endif
}

bool
Gui::progress_visible(int state)
{
    if (state < 0)
	return _progress_visible;
    else
	return (_progress_visible = state);
}

bool
Gui::command(string command)
{
    if (!show_gui())		// Bail if were not supposed to be here...
	return true;

    DBG(cerr << "Sending the GUI: " << command << endl);

#if HAVE_EXPECT

    if (_gui < 0) {
	bool running = _start_gui();
	if (!running)
	    return false;
    }
    
    unsigned status = write(_gui, command.c_str(), command.length());
    if (status != command.length()) {
	cerr << "Wrong number of chars transferred to GUI!" << endl;
	_user_wants_gui = false;
	return false;
    }
    
    exp_timeout = 10;		// non-modal interactions should not take long.

    string response = (string)".*OK\r\n" + GUI_PROMPT;
    string error = (string)".*(error)|(unknown)|(invalid).*\r\n" + GUI_PROMPT;
    status = exp_expectl(_gui, 
			 exp_regexp, response.c_str(), OK, 
			 exp_regexp, error.c_str(), ERROR,
			 exp_end);

    switch (status) {
      case OK:
	status = true;
	break;

      case ERROR:
	cerr << "Expect reports an error" << endl;
	status = false;
	break;

      case EXP_TIMEOUT:
	cerr << "Expect timeout" << endl;
	status = false;
	break;

      case EXP_EOF:
	cerr << "Expect found unexpectedly EOF" << endl;
	status = false;
	break;

      case EXP_FULLBUFFER:
	cerr << "Expect buffer filled" << endl;
	status = false;
	break;

      default:
	cerr << "Unknown error from expect" << endl;
	status = false;
	break;
    }
    
    if (!status) {
	cerr << "GUI Could not execute command: " << command << endl
	     << "Shutting down the GUI." << endl;
	if (kill(_gui_pid, SIGKILL) == 0)
	    waitpid(_gui_pid, NULL, 0);	// only wait if kill works
	_show_gui = false;
	_gui = -1;		// Memory leak...
    }

    return status;

#else // !HAVE_EXPECT

    return true;

#endif
}

// Put the result read from the interpreter in RESULT. If this function
// returns false, the contents of RESULT are undefined.

bool
Gui::response(string command, string &result)
{
    // Unlike send_gui_command(), return false if we try to get a value from
    // the gui when it is not being used.
    if (!show_gui())
	return false;

    DBG(cerr << "Sennding GUI: " << command << endl);

#if HAVE_EXPECT

    if (_gui < 0) {
	bool running = _start_gui();
	if (!running)
	    return false;
    }
    
    unsigned status = write(_gui, command.c_str(), command.length());
    if (status != command.length()) {
	cerr << "Wrong number of chars transferred to GUI!" << endl;
	_show_gui = false;
	return false;
    }
    
    // For modal interactions wait for the response.
    exp_timeout = -1;

    // the Response from the GUI is always either "OK" followed by some
    // information or "CANCEL" followed by nothing.
    string response = (string)"*OK\r\n" + GUI_PROMPT;
    string cancel = (string)"*CANCEL\r\n" + GUI_PROMPT;
    string error = (string)".*(unknown)|(invalid).*\r\n" + GUI_PROMPT;
    status = exp_expectl(_gui, 
			 exp_glob, response.c_str(), OK, 
			 exp_glob, cancel.c_str(), CANCEL,
			 exp_regexp, error.c_str(), ERROR,
			 exp_end);

    switch (status) {
      case OK: {
	// Slurp up the rest into RESULT.
	// The buffer used by expect looks like:
	// **********************
	// |   |       |         |
	// |   |       |         exp_buffer_end (one past last char)
	// |   |       exp_match_end (one char past last matched char)
	// |   exp_match (first char)
	// exp_buffer (first char)
	//
	// To get the result we want the stuff after the matched text
	int len = exp_buffer_end - exp_match_end;
	char *tmp = new char [len + 1];
	memcpy(tmp, exp_match_end, len);
	tmp[len] = '\0';
	result = tmp;
	delete tmp;
	status = true;
	break;
      }

      case CANCEL:
	result = "";
	status = true;
	break;
	
      case ERROR:
      case EXP_TIMEOUT:
      case EXP_EOF:
      case EXP_FULLBUFFER:
      default:
	cerr << "GUI Could not execute command: " << command << endl
	     << "Shutting down the GUI." << endl;
	DBG(cerr << "Contents of the expect buffer: " << exp_buffer << endl);
	if (kill(_gui_pid, SIGKILL) == 0)
	    waitpid(_gui_pid, NULL, 0);
	_show_gui = false;
	_gui = -1;
	result = "";
	status = false;
	break;
    }

    return status;

#else // !HAVE_EXPECT

    return false;

#endif
}

bool
Gui::show_gui(int state)
{
    if (state < 0)
	return _show_gui && _user_wants_gui;
    else
	return (_show_gui = (bool)state) && _user_wants_gui;
}

